/*
   Filename:     dinit.c
   Version:      0.1
   Author:       Peter W"achtler

   Copying:     Copyright 2000 Peter W"achtler (pwaechtler@qnx.de)
 
   Portions copyright 1993, 1994 David Hudson (dave@humbug.demon.co.uk)
   Portions copyright 1992, 1993 Remy Card (card@masi.ibp.fr)
   and 1991 Linus Torvalds (torvalds@klaava.helsinki.fi)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <fcntl.h>
#include <linux/hdreg.h>
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/fd.h>
#include <mntent.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <endian.h>
#include <ctype.h>

#include <linux/qnx4_fs.h>

#define TEST_BUFFER_BLOCKS 16
#define SECTORS_PER_BLOCK ( BLOCK_SIZE / SECTOR_SIZE )

/* Report a failure message and return a failure error code */

#define die( str ) fatal_error( "%s: " str "\n" )

#if BIG_ENDIAN == BYTE_ORDER

#define CF_LE_W(v) ((((v) & 0x00ff) << 8) | (((v) & 0xff00) >> 8))
#define CT_LE_W(v) ((((v) & 0x00ff) << 8) | (((v) & 0xff00) >> 8))
#define CF_LE_L(v) ((((v) & 0x0000ff00) <<  8) | (((v) & 0x00ff0000) >> 8) | \
                    (((v) & 0x000000ff) << 24) | (((v) & 0xff000000) >> 24))
#define CT_LE_L(v) ((((v) & 0x0000ff00) <<  8) | (((v) & 0x00ff0000) >> 8) | \
                    (((v) & 0x000000ff) << 24) | (((v) & 0xff000000) >> 24))

#else

#define CF_LE_W(v) v
#define CF_LE_L(v) v
#define CT_LE_W(v) v
#define CT_LE_L(v) v

#endif

#include "ext2fs.h"

errcode_t ext2fs_get_device_size(const char *file, int blocksize,
                                 blk_t *retblocks);

#define SECTOR_SIZE     512 /* sector size (bytes) */

#define BOOT_SIGN 0xAA55	/* Boot sector magic number */

char bootsect[]={
'\xEB','\x4A','\x90','\x48','\x51','\x4E','\x58','\x34',
'\x46','\x53','\x20','\x00','\x02','\x01','\xEF','\xFF',
'\x01','\x40','\x00','\x00','\x00','\xF8','\x10','\x00',
'\x3F','\x00','\xFF','\x00','\x55','\x97','\x13','\x00',
'\x54','\x97','\x14','\x00','\x80','\x00','\x29','\x93',
'\xEC','\x09','\x38','\x50','\x72','\x65','\x73','\x73',
'\x20','\x45','\x73','\x63','\x20','\x66','\x6F','\x72',
'\x20','\x61','\x6C','\x74','\x65','\x72','\x6E','\x61',
'\x74','\x65','\x20','\x4F','\x53','\x00','\x00','\x18',
'\x7B','\xF8','\x8B','\x00','\xE8','\x00','\x00','\xCD',
'\x12','\x83','\xE8','\x04','\xC1','\xE0','\x06','\x0E',
'\x1F','\x5E','\x83','\xEE','\x4F','\x8E','\xC0','\x31',
'\xFF','\xB9','\x00','\x01','\xFA','\xF3','\xA5','\xFB',
'\x06','\x68','\x6D','\x00','\xCB','\x0E','\x1F','\x0E',
'\x17','\xBC','\x00','\x08','\xBE','\x2B','\x00','\xAC',
'\x08','\xC0','\x74','\x05','\xE8','\x75','\x01','\xEB',
'\xF6','\xA1','\x1A','\x00','\xF6','\x26','\x18','\x00',
'\xA3','\x27','\x00','\xB8','\x02','\x00','\x99','\xBF',
'\x01','\x00','\xBB','\x00','\x02','\xE8','\xC0','\x00',
'\x8E','\xD9','\xB9','\x24','\x00','\xB8','\x00','\x01',
'\xCD','\x16','\x75','\x0C','\x8B','\x16','\x6C','\x04',
'\x3B','\x16','\x6C','\x04','\x74','\xFA','\xE2','\xED',
'\x0E','\x1F','\xBE','\x80','\x02','\x3C','\x1B','\x75',
'\x0E','\x80','\x0E','\x03','\x00','\x20','\x83','\x7C',
'\x70','\x00','\x74','\x03','\x83','\xC6','\x40','\x8B',
'\x4C','\x30','\x8B','\x44','\x14','\x8B','\x54','\x16',
'\xBB','\x00','\x04','\xE8','\x82','\x00','\x8B','\x2E',
'\x0A','\x04','\x89','\x2E','\x29','\x00','\x8B','\x44',
'\x1C','\x8B','\x54','\x1E','\xA3','\x00','\x04','\x89',
'\x16','\x02','\x04','\xC6','\x06','\x08','\x04','\x01',
'\x83','\xC6','\x14','\xEB','\x1E','\x83','\xC6','\x08',
'\xFE','\x0E','\x08','\x04','\x75','\x15','\xA1','\x00',
'\x04','\x8B','\x16','\x02','\x04','\x1E','\x07','\xBB',
'\x00','\x04','\xBF','\x01','\x00','\xE8','\x48','\x00',
'\xBE','\x10','\x04','\x8B','\x04','\x8B','\x54','\x02',
'\x8B','\x7C','\x04','\x8E','\xC5','\x31','\xDB','\xE8',
'\x36','\x00','\x01','\xF8','\x83','\xD2','\x00','\x29',
'\x7C','\x04','\xC1','\xE7','\x05','\x01','\xFD','\x83',
'\x7C','\x04','\x00','\x75','\xE3','\xE2','\xBE','\xB0',
'\x0D','\xE8','\xB8','\x00','\xB0','\x0A','\xE8','\xB3',
'\x00','\xB0','\xFF','\xE6','\x21','\xFA','\x8A','\x1E',
'\x03','\x00','\x8E','\x06','\x29','\x00','\x26','\x88',
'\x1E','\x0C','\x00','\x06','\x31','\xC0','\x50','\xCB',
'\x50','\x53','\x51','\x52','\x83','\xE8','\x01','\x83',
'\xDA','\x00','\x03','\x06','\x48','\x00','\x13','\x16',
'\x4A','\x00','\x50','\x52','\x4F','\x01','\xF8','\x83',
'\xD2','\x00','\x47','\xF7','\x36','\x27','\x00','\x3D',
'\x00','\x04','\x72','\x07','\xB0','\x43','\xE8','\x73',
'\x00','\xEB','\xFE','\x5A','\x58','\xF7','\x36','\x27',
'\x00','\x88','\xC5','\x88','\xE1','\xC0','\xC9','\x02',
'\x89','\xD0','\x99','\xF6','\x36','\x18','\x00','\x88',
'\xC6','\x08','\xE1','\xFE','\xC1','\x8A','\x16','\x24',
'\x00','\xF6','\xC2','\x80','\x74','\x04','\x88','\xD0',
'\xEB','\x05','\xA0','\x18','\x00','\x28','\xE0','\x30',
'\xE4','\x39','\xF8','\x7F','\x02','\x89','\xC7','\x89',
'\xF8','\xB4','\x02','\xCD','\x13','\x73','\x22','\x80',
'\xFC','\x09','\x75','\x0D','\x8C','\xC7','\xF7','\xDF',
'\x81','\xE7','\xFF','\x0F','\xC1','\xEF','\x05','\xEB',
'\xE6','\xB0','\x3F','\xE8','\x1E','\x00','\x31','\xC0',
'\xBF','\x01','\x00','\xCD','\x13','\x73','\xD8','\x72',
'\xFE','\xF6','\xC2','\x80','\x75','\x04','\x30','\xE4',
'\x89','\xC7','\xB0','\x2E','\xE8','\x05','\x00','\x5A',
'\x59','\x5B','\x58','\xC3','\x53','\xBB','\x07','\x00',
'\xB4','\x0E','\xCD','\x10','\x5B','\xC3','\x55','\xAA',
};

typedef struct __tag_rootblock {
	struct qnx4_inode_entry root;
	struct qnx4_inode_entry inode;
	struct qnx4_inode_entry boot;
	struct qnx4_inode_entry altboot;
	struct qnx4_inode_entry lostfound;  /* lost+found? */
	char filler[512-(5*sizeof(struct qnx4_inode_entry))];
} qnx4_rootblock_t;

/* a direntry should be a union of qnx4_inode_entry and qnx4_link_info
	if qnx4_inode_entry.status has QNX4_LINK_FILE bit set it is
	a qnx4_link_info 
*/
typedef struct __tag_rootdir {
	struct qnx4_link_info dot;
	struct qnx4_link_info dotdot;
	struct qnx4_inode_entry bitmap;
	struct qnx4_link_info inodes;
	struct qnx4_link_info boot;
	struct qnx4_link_info altboot;
	struct qnx4_link_info lostfound;
	struct qnx4_inode_entry zero;
} qnx4_rootdir_t;

typedef struct __tag_lostfound {
	struct qnx4_link_info dot;
	struct qnx4_link_info dotdot;
	struct qnx4_link_info zero;
} qnx4_lostfound_t;

typedef struct __tag_inodefile {
	struct qnx4_link_info signature;
	struct qnx4_link_info zero;
} qnx4_inodefile_t;

/* Global variables - the root of all evil :-) - see these and weep! */

qnx4_rootblock_t rootblock;
qnx4_rootdir_t rootdir;
qnx4_lostfound_t lostfound;
qnx4_inodefile_t inode;

/* in the "." entry in a directory the size of the extent 
	is saved for crash recovery */
#define DIRECTORY_SIGNATURE_OFFSET 8
static char dirsig[]={'.',0,'I',3,'Q','N','X',0,8,0,0,0,0,0,0,0};
static char inodesig[]="IamTHE.inodeFILE";
#define FSID_OFFSET 3

static int bitmap_size;
static int bitmap_blocks;
static int used_blocks;
static int rest,bits;
static int boot=0;
static int harddisk=0;
static unsigned char reststring[2];

static char *program_name = "dinit";	/* Name of the program */
static char *device_name = NULL;	/* Name of the device on which to create the filesystem */
static int check = 0;	/* Default to no readablity checking */
static int verbose = 0;	/* Default to verbose mode off */
static time_t create_time;	/* Creation time */

static blk_t blocks;		/* Number of blocks in filesystem */
static int badblocks = 0;	/* Number of bad blocks in the filesystam */
static int dev = -1;		/* FS block device file handle */
static unsigned int currently_testing = 0;	/* Block currently being tested (if autodetect bad blocks) */
static int start_data_block;	/* Block number for the start of the data area */
static int block_size = BLOCK_SIZE;	/* Size of a block */


static volatile void fatal_error (const char *fmt_string);
static long do_check (char *buffer, int try, unsigned int current_block);
static void alarm_intr (int alnum);
static void check_blocks (void);
static void get_list_blocks (char *filename);
static void check_mount (char *device_name);


/* Handle the reporting of fatal errors.  Volatile to let gcc know that this doesn't return */
static volatile void
fatal_error (const char *fmt_string)
{
  fprintf (stderr, fmt_string, program_name, device_name);
  exit (1);			/* The error exit code is 1! */
}

/* Perform a test on a block.  Return the number of blocks that could be read successfully */
static long
do_check (char *buffer, int try, unsigned int current_block)
{
  long got;

  if (lseek (dev, current_block * block_size, SEEK_SET)		/* Seek to the correct location */
      != current_block * block_size)
    die ("seek failed during testing for blocks");

  got = read (dev, buffer, try * block_size);	/* Try reading! */
  if (got < 0)
    got = 0;

  if (got & (block_size - 1))
    printf ("Unexpected values in do_check: probably bugs\n");
  got /= block_size;

  return got;
}


/* Alarm clock handler - display the status of the quest for bad blocks!  Then retrigger the alarm for 
   five seconds later (so we can come here again) */

static void
alarm_intr (int alnum)
{
  if (currently_testing >= blocks)
    return;

  signal (SIGALRM, alarm_intr);
  alarm (5);
  if (!currently_testing)
    return;

  printf ("%d... ", currently_testing);
  fflush (stdout);
}


static void
check_blocks (void)
{
  int try, got;
  int i;
  static char blkbuf[BLOCK_SIZE * TEST_BUFFER_BLOCKS];

  if (verbose) {
      printf ("Searching for bad blocks ");
      fflush (stdout);
  }
  currently_testing = 0;
  if (verbose) {
      signal (SIGALRM, alarm_intr);
      alarm (5);
  }
  try = TEST_BUFFER_BLOCKS;
  while (currently_testing < blocks) {
      if (currently_testing + try > blocks)
	try = blocks - currently_testing;
      got = do_check (blkbuf, try, currently_testing);
      currently_testing += got;
      if (got == try)
	{
	  try = TEST_BUFFER_BLOCKS;
	  continue;
	}
      else
	try = 1;
      if (currently_testing < start_data_block)
	die ("bad blocks before data-area: cannot make fs");

      for (i = 0; i < SECTORS_PER_BLOCK; i++)	/* Mark all of the sectors in the block as bad */
//	mark_sector_bad (currently_testing * SECTORS_PER_BLOCK + i);
      badblocks++;
      currently_testing++;
  }

  if (verbose)
    printf ("\n");

  if (badblocks)
    printf ("%d bad block%s\n", badblocks,
	    (badblocks > 1) ? "s" : "");
}


static void
get_list_blocks (char *filename)
{
  int i;
  FILE *listfile;
  unsigned long blockno;

  listfile = fopen (filename, "r");
  if (listfile == (FILE *) NULL)
    die ("Can't open file of bad blocks");

  while (!feof (listfile))
    {
      fscanf (listfile, "%ld\n", &blockno);
      for (i = 0; i < SECTORS_PER_BLOCK; i++)	/* Mark all of the sectors in the block as bad */
//	mark_sector_bad (blockno * SECTORS_PER_BLOCK + i);
      badblocks++;
    }
  fclose (listfile);

  if (badblocks)
    printf ("%d bad block%s\n", badblocks,
	    (badblocks > 1) ? "s" : "");
}


/* Check to see if the specified device is currently mounted - abort if it is */
static void
check_mount (char *device_name)
{
  FILE *f;
  struct mntent *mnt;

  if ((f = setmntent (MOUNTED, "r")) == NULL)
    return;
  while ((mnt = getmntent (f)) != NULL)
    if (strcmp (device_name, mnt->mnt_fsname) == 0)
      die ("%s contains a mounted file system.");
  endmntent (f);
}

void
usage (void)
{
  fatal_error("\
Usage: %s [OPTIONS] /dev/name [blocks]\n\n\
  -b            install boot sector code\n\
  -h            device is a hard disk\n\
  -c            Check filesystem as it gets built (not implemented yet)\n\
  -l <filename> Bad block filename (not implemented yet)\n\
  -v            Verbose execution\n");
}


int
main (int argc, char **argv)
{
  int c;
  char *tmp;
  char *listfile = NULL;
  struct stat statbuf;
  int i;

  if (argc && *argv)		/* What's the program name? */
    program_name = *argv;

  time(&create_time);

  fprintf (stderr, "%s 0.1, 27th January 2000 for QNX4 FS\n",
	   program_name);

  while ((c = getopt (argc, argv, "bhcl:v")) != EOF){
    switch (c) {
      case 'c':		/* c : Check FS as we build it */
		check = 1;
		break;
      case 'l':		/* l : Bad block filename */
		listfile = optarg;
		break;
      case 'v':		/* v : Verbose execution */
		verbose = 1;
		break;
	  case 'b':
	  	boot= 1;
		break;
	  case 'h':
	  	harddisk= 1;
		break;
      default:
		usage ();
    }
  }
  device_name = argv[optind];	/* Determine the number of blocks in the FS */
  if (optind == argc - 2)	/*  Either check the user specified number */
    blocks = strtol (argv[optind + 1], &tmp, 0);
  else if (optind == argc - 1)	/*  Or have a look and see! */ {
      if (0==ext2fs_get_device_size(device_name,512,&blocks))
		  fprintf(stderr,"size of %s: %d blocks (512 bytes)\n",device_name,blocks);
	  else
	  	fprintf(stderr,"could not determine size of blockdevice %s\n",device_name);
      tmp = "";
  } else
    usage ();
  if (*tmp) {
      printf ("Bad block count : %s\n", argv[optind + 1]);
      usage ();
  }

  if (check && listfile)/* Auto and specified bad block handling are mutually */
    die ("-c and -l are incompatible");		/* exclusive of each other! */

  check_mount (device_name);	/* Is the device already mounted? */

  if ((dev = open (device_name, O_RDWR))<0)
    die ("unable to open %s");
  if (fstat (dev, &statbuf) < 0)
    die ("unable to stat %s");
  if (!S_ISBLK (statbuf.st_mode))
    check = 0;
  else
    /* Ignore any 'full' fixed disk devices */
    if ((statbuf.st_rdev & 0xff3f) == 0x0300 || /* hda, hdb */
	(statbuf.st_rdev & 0xff0f) == 0x0800 || /* sd */
	(statbuf.st_rdev & 0xff3f) == 0x0d00 || /* xd */
	(statbuf.st_rdev & 0xff3f) == 0x1600 )  /* hdc, hdd */
      die ("Will not try to make filesystem on '%s'");


  if (check)		/* Determine any bad block locations and mark them */
    check_blocks ();
  else if (listfile)
    get_list_blocks (listfile);

  /* determine size of bitmap  */
  bitmap_size= (blocks >>3);	
  bitmap_blocks=(bitmap_size>>9)+1;

  /* zeroed block for deleting */  
  if (!(tmp=calloc(BLOCK_SIZE,1)))
  	die("Out of memory");

  /* create rootblock */
  rootblock.root.di_fname[0]='/';
  rootblock.root.di_size=CT_LE_L(harddisk? 0x1000:0x800);
  rootblock.root.di_first_xtnt.xtnt_blk= CT_LE_L(bitmap_blocks+3);
  rootblock.root.di_first_xtnt.xtnt_size=CT_LE_L(harddisk? 8:4);
  rootblock.root.di_xblk=0;
  rootblock.root.di_ftime=CT_LE_L(create_time);
  rootblock.root.di_mtime=CT_LE_L(create_time);
  rootblock.root.di_atime=CT_LE_L(create_time);
  rootblock.root.di_ctime=CT_LE_L(create_time);
  rootblock.root.di_num_xtnts=CT_LE_W(1);
  rootblock.root.di_mode=CT_LE_W(0x41fd);
  rootblock.root.di_uid=0;
  rootblock.root.di_gid=0;
  rootblock.root.di_nlink=CT_LE_W(3);
  rootblock.root.di_type=0;
  rootblock.root.di_status=QNX4_FILE_FSYSCLEAN|QNX4_FILE_USED;

  rootblock.inode.di_fname[0]=0;
  rootblock.inode.di_size=CT_LE_W(harddisk? 0x2000:0x200);
  rootblock.inode.di_first_xtnt.xtnt_blk= CT_LE_L(
    rootblock.root.di_first_xtnt.xtnt_blk+
	rootblock.root.di_first_xtnt.xtnt_size);
  rootblock.inode.di_first_xtnt.xtnt_size=CT_LE_L(harddisk? 16:1);
  rootblock.inode.di_xblk=0;
  rootblock.inode.di_ftime=CT_LE_L(create_time);
  rootblock.inode.di_mtime=CT_LE_L(create_time);
  rootblock.inode.di_atime=CT_LE_L(create_time);
  rootblock.inode.di_ctime=CT_LE_L(create_time);
  rootblock.inode.di_num_xtnts=CT_LE_W(1);
  rootblock.inode.di_mode=CT_LE_W(0x8124);
  rootblock.inode.di_uid=0;
  rootblock.inode.di_gid=0;
  rootblock.inode.di_nlink=CT_LE_W(1);
  rootblock.inode.di_type=0;
  rootblock.inode.di_status=QNX4_FILE_INODE|QNX4_FILE_USED;

  rootblock.boot.di_fname[0]=0;
  rootblock.boot.di_size=0x0;	/* empty bootfile */
  rootblock.boot.di_first_xtnt.xtnt_blk= 0;
  rootblock.boot.di_first_xtnt.xtnt_size= 0;
  rootblock.boot.di_xblk=0;
  rootblock.boot.di_ftime=CT_LE_L(create_time);
  rootblock.boot.di_mtime=CT_LE_L(create_time);
  rootblock.boot.di_atime=CT_LE_L(create_time);
  rootblock.boot.di_ctime=CT_LE_L(create_time);
  rootblock.boot.di_num_xtnts=0;
  rootblock.boot.di_mode=CT_LE_W(0x8180);
  rootblock.boot.di_uid=0;
  rootblock.boot.di_gid=0;
  rootblock.boot.di_nlink=CT_LE_W(1);
  rootblock.boot.di_type=0;
  rootblock.boot.di_status=QNX4_FILE_INODE|QNX4_FILE_USED;

  rootblock.altboot.di_fname[0]=0;
  rootblock.altboot.di_size=0x0;	/* empty alternate bootfile */
  rootblock.altboot.di_first_xtnt.xtnt_blk= 0;
  rootblock.altboot.di_first_xtnt.xtnt_size= 0;
  rootblock.altboot.di_xblk=0;
  rootblock.altboot.di_ftime=CT_LE_L(create_time);
  rootblock.altboot.di_mtime=CT_LE_L(create_time);
  rootblock.altboot.di_atime=CT_LE_L(create_time);
  rootblock.altboot.di_ctime=CT_LE_L(create_time);
  rootblock.altboot.di_num_xtnts=0;
  rootblock.altboot.di_mode=CT_LE_W(0x8180);
  rootblock.altboot.di_uid=0;
  rootblock.altboot.di_gid=0;
  rootblock.altboot.di_nlink=CT_LE_W(1);
  rootblock.altboot.di_type=0;
  rootblock.altboot.di_status=QNX4_FILE_INODE|QNX4_FILE_USED;

  if (harddisk){
  	/* is it a lost+found, or what? docs says something about a label */
	  rootblock.lostfound.di_fname[0]=0;
	  rootblock.lostfound.di_size=0x2000;	
	  rootblock.lostfound.di_first_xtnt.xtnt_blk= CT_LE_L(
    	rootblock.inode.di_first_xtnt.xtnt_blk+16);
	  rootblock.lostfound.di_first_xtnt.xtnt_size=CT_LE_L(16);
	  rootblock.lostfound.di_xblk=0;
	  rootblock.lostfound.di_ftime=CT_LE_L(create_time);
	  rootblock.lostfound.di_mtime=CT_LE_L(create_time);
	  rootblock.lostfound.di_atime=CT_LE_L(create_time);
	  rootblock.lostfound.di_ctime=CT_LE_L(create_time);
	  rootblock.lostfound.di_num_xtnts=CT_LE_W(1);
	  rootblock.lostfound.di_mode=CT_LE_W(0x41ed);
	  rootblock.lostfound.di_uid=0;
	  rootblock.lostfound.di_gid=0;
	  rootblock.lostfound.di_nlink=CT_LE_W(2);
	  rootblock.lostfound.di_type=0;
	  rootblock.lostfound.di_status=QNX4_FILE_INODE|QNX4_FILE_USED;
  }
  /* create root-dir */
  /* a link into the rootblock */
  memcpy(rootdir.dot.dl_fname,dirsig,sizeof(dirsig));
  rootdir.dot.dl_fname[DIRECTORY_SIGNATURE_OFFSET]=
  	CT_LE_L(rootblock.root.di_first_xtnt.xtnt_size);
  rootdir.dot.dl_inode_blk=CT_LE_L(2);
  rootdir.dot.dl_inode_ndx=0;
  rootdir.dot.dl_status=QNX4_FILE_LINK;
  
  /* a link into the rootblock */
  strcpy(rootdir.dotdot.dl_fname,"..");
  rootdir.dotdot.dl_inode_blk=CT_LE_L(2);
  rootdir.dotdot.dl_inode_ndx=0;
  rootdir.dotdot.dl_status=QNX4_FILE_LINK;

  /* full inode entry */
  strcpy(rootdir.bitmap.di_fname,".bitmap");	/* FIXME */
  rootdir.bitmap.di_size=CT_LE_L(bitmap_size);
  rootdir.bitmap.di_first_xtnt.xtnt_blk= CT_LE_L(3);
  rootdir.bitmap.di_first_xtnt.xtnt_size= CT_LE_L(bitmap_blocks);
  rootdir.bitmap.di_xblk=0;
  rootdir.bitmap.di_ftime=CT_LE_L(create_time);
  rootdir.bitmap.di_mtime=CT_LE_L(create_time);
  rootdir.bitmap.di_atime=CT_LE_L(create_time);
  rootdir.bitmap.di_ctime=CT_LE_L(create_time);
  rootdir.bitmap.di_num_xtnts=CT_LE_W(1);
  rootdir.bitmap.di_mode=CT_LE_W(0x8124);
  rootdir.bitmap.di_uid=0;
  rootdir.bitmap.di_gid=0;
  rootdir.bitmap.di_nlink=CT_LE_W(1);
  rootdir.bitmap.di_type=0;
  rootdir.bitmap.di_status=QNX4_FILE_USED; /* hmh, no inode? */
  
  /* a link into the rootblock */
  strcpy(rootdir.inodes.dl_fname,".inodes");
  rootdir.inodes.dl_inode_blk=CT_LE_L(2);
  rootdir.inodes.dl_inode_ndx=1;
  rootdir.inodes.dl_status=QNX4_FILE_LINK;

  /* a link into the rootblock */
  strcpy(rootdir.boot.dl_fname,".boot");
  rootdir.boot.dl_inode_blk=CT_LE_L(2);
  rootdir.boot.dl_inode_ndx=2;
  rootdir.boot.dl_status=QNX4_FILE_LINK;
  
  /* a link into the rootblock */
  strcpy(rootdir.altboot.dl_fname,".altboot");
  rootdir.altboot.dl_inode_blk=CT_LE_L(2);
  rootdir.altboot.dl_inode_ndx=3;
  rootdir.altboot.dl_status=QNX4_FILE_LINK;
  
  /* what's this? not marked as inode nor link - instead busy and mysterious spare[1] */
  /* could be a "directory signature extent length" */
  if (harddisk){
	  strcpy(rootdir.lostfound.dl_fname,"lost+found");
	  rootdir.lostfound.dl_inode_blk=CT_LE_L(2);
	  rootdir.lostfound.dl_inode_ndx=4;
	  rootdir.lostfound.dl_spare[1]=8;   /* meaning? initial size is 8192(16 blocks) */
	  rootdir.lostfound.dl_status=QNX4_FILE_BUSY;

	  /* create lost+found */
	  memcpy(lostfound.dot.dl_fname,dirsig,sizeof(dirsig));
	  lostfound.dot.dl_fname[DIRECTORY_SIGNATURE_OFFSET]=
  		CT_LE_L(rootblock.lostfound.di_first_xtnt.xtnt_size);
	  lostfound.dot.dl_inode_blk=CT_LE_L(2);
	  lostfound.dot.dl_inode_ndx=4;
	  lostfound.dot.dl_status=QNX4_FILE_LINK;

	  strcpy(lostfound.dotdot.dl_fname,"..");
	  lostfound.dotdot.dl_inode_blk=
    	  CT_LE_L(rootblock.root.di_first_xtnt.xtnt_blk);
	  lostfound.dotdot.dl_inode_ndx=0;
	  lostfound.dotdot.dl_status=QNX4_FILE_LINK;
  }
  
  /* create .inode */
  strcpy(inode.signature.dl_fname,inodesig);
  
  /* check all meta-data for bad blocks */	/* FIXME */
  
  /* write everything including boot sector */
  if (harddisk)
  	bootsect[FSID_OFFSET]='H';
  else
  	bootsect[FSID_OFFSET]='F';
  if (boot){
	/* write boot sector */
	lseek(dev,0,SEEK_SET);
	write(dev,bootsect,sizeof(bootsect));
  }
  /* write root block */  
  lseek(dev,512,SEEK_SET);
  write(dev,&rootblock,sizeof(rootblock));

  /* write .bitmap */
  /* init all bitmap blocks to zero */
  lseek(dev,1024,SEEK_SET);
  for (i=0;i<bitmap_blocks;i++)	/* FIXME */
  	write(dev,tmp,BLOCK_SIZE);

  /* mark the first used blocks */
  used_blocks=2+bitmap_blocks+rootblock.root.di_first_xtnt.xtnt_size+
  	rootblock.inode.di_first_xtnt.xtnt_size+
  	rootblock.lostfound.di_first_xtnt.xtnt_size;

  lseek(dev,1024,SEEK_SET);
  for (i=0;i<(used_blocks>>4);i++)	/* all 16 blocks used */
  	write(dev,"\xff\xff",2);
  rest=(used_blocks%16);	/* bitmap is organized in Intel LE 16bit-words */
  for (bits=0,i=0;i<rest;i++)
  	bits= (bits<<1)+1;
  rest=CT_LE_L(bits);
  reststring[0]=(rest&0xff);
  reststring[1]=(rest>>8)&0xff;
  write(dev,reststring,2);

  /* mark the bits behind regular bitmap as used */  
  lseek(dev,1024+bitmap_size,SEEK_SET);
  for (i=(bitmap_blocks<<9)-bitmap_size;i;i--)	/* FIXME */
  	write(dev,"\xff",1);

  /* write root dir */  
  lseek(dev,(rootblock.root.di_first_xtnt.xtnt_blk-1)<<9,SEEK_SET);
  write(dev,&rootdir,sizeof(rootdir));
  for (i=0;i<rootblock.root.di_first_xtnt.xtnt_size-1;i++)
  	write(dev,tmp,BLOCK_SIZE);	/* zero out rest */
  
  /* write inode file */  
  lseek(dev,(rootblock.inode.di_first_xtnt.xtnt_blk-1)<<9,SEEK_SET);
  write(dev,&inode,sizeof(inode));
  for (i=0;i<rootblock.inode.di_first_xtnt.xtnt_size-1;i++)
  	write(dev,tmp,BLOCK_SIZE);	/* zero out rest */
  
  /* write lost+found dir */  
  if (harddisk){
	  lseek(dev,(rootblock.lostfound.di_first_xtnt.xtnt_blk-1)<<9,SEEK_SET);
	  write(dev,&lostfound,sizeof(lostfound));
	  for (i=0;i<rootblock.lostfound.di_first_xtnt.xtnt_size-1;i++)
  		write(dev,tmp,BLOCK_SIZE);	/* zero out rest */
  }
  exit (0);			/* Terminate with no errors! */
}
